#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_multimin.h>

struct data {
  size_t m;
  size_t n;
  double *X;
  double *y;
};

double
my_f (const gsl_vector *v, void *params)
{

    size_t n = ((struct data *)params)->n;
    size_t m = ((struct data *)params)->m;
    double *X = ((struct data *)params)->X;
    double *y = ((struct data *)params)->y;

    double res = 0.0;
    for (size_t i = 0; i < m; i++) {
        double tmp = 0.0;
        for (size_t j = 0; j<n; j++) {
            double pj = gsl_vector_get(v,j);
            tmp += X[i*n+j]*pj;
        }
        tmp = pow(tmp,2);
        res += pow(tmp-y[i],2);
    }


  return res;
}

/* The gradient of f */
void
my_df (const gsl_vector *v, void *params,
       gsl_vector *df)
{
    size_t n = ((struct data *)params)->n;

    double fval = my_f(v, params);
    double h = 1e-7;

    gsl_vector *x= gsl_vector_alloc(n);
    gsl_vector_memcpy(x, v);
    /* Forward Finite differences*/
    for (size_t i=0; i<n; i++){
        double tmp = gsl_vector_get(v,i);
        gsl_vector_set(x,i,tmp+h);
        double fval_h = my_f(x,params);
        double deriv = (fval_h-fval)/h;
        gsl_vector_set(df,i, deriv);
        gsl_vector_set(x,i,tmp);
    }
    gsl_vector_free(x);
}

/* Compute both f and df together. */
void
my_fdf (const gsl_vector *x, void *params,
        double *f, gsl_vector *df)
{
  *f = my_f(x, params);
  my_df(x, params, df);
}

int
main (int argc, char* argv[])
{
  assert(argc==3); size_t m=std::stoi(argv[1]), n=std::stoi(argv[2]);

  size_t iter = 0;
  int status;

  const gsl_multimin_fdfminimizer_type *T;
  gsl_multimin_fdfminimizer *s;

  Eigen::Matrix<double, Eigen::Dynamic, 1> y;
  Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> t;
  /* this is the data to be fitted */
  t = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(m,n);
  y = Eigen::Matrix<double, Eigen::Dynamic, 1>::Random(m);
  t=t.cwiseProduct(t);

  double *Y = new double[m];
  for (size_t i=0; i<m; i++) Y[i] = y(i);

  double *tt = new double[m*n];
  for (size_t i=0; i<m; i++){
      for (size_t j=0; j<n; j++)
          tt[i*n+j] = t(i,j);
  }

  struct data par = {m, n, tt, Y};

  gsl_multimin_function_fdf my_func;

  my_func.n = n;
  my_func.f = my_f;
  my_func.df = my_df;
  my_func.fdf = my_fdf;
  my_func.params = &par;

  gsl_vector *x;
  x = gsl_vector_alloc (n);
  for (size_t i=0; i<n; i++) gsl_vector_set(x, i, -1.0);

  /* Choose the minization algorithm */
  T = gsl_multimin_fdfminimizer_conjugate_fr;
  s = gsl_multimin_fdfminimizer_alloc (T, n);

  gsl_multimin_fdfminimizer_set (s, &my_func, x, 0.01, 1e-4);

  printf("Iteration   Function Value          Current  Solution\n");
  do
    {
      iter++;
      status = gsl_multimin_fdfminimizer_iterate (s);

      if (status)
        break;

      status = gsl_multimin_test_gradient (s->gradient, 1e-5);

      if (status == GSL_SUCCESS)
        printf ("Minimum found at:\n");

      printf ("%5ld       %10.5f", iter,
              s->f);
      size_t max_out = n;
      if (max_out>5) max_out=5;

      for (size_t i=0; i<max_out; i++){
          printf ("          %.5f",
                  gsl_vector_get (s->x, i));
      }
      printf("\n");
    }
  while (status == GSL_CONTINUE && iter < 100);

  gsl_multimin_fdfminimizer_free (s);
  gsl_vector_free (x);

  delete [] tt;
  delete [] Y;

  return 0;
}
