#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlinear.h>

struct data {
  size_t m;
  size_t n;
  double *X;
  double *y;
};

int
expb_f (const gsl_vector * p, void *par,
        gsl_vector * f)
{
  size_t n = ((struct data *)par)->n;
  size_t m = ((struct data *)par)->m;
  double *X = ((struct data *)par)->X;
  double *y = ((struct data *)par)->y;

  for (size_t i = 0; i < m; i++) {
      double tmp = 0.0;
      for (size_t j = 0; j<n; j++) {
          double pj = gsl_vector_get(p,j);
          tmp += X[i*n+j]*pj;
      }
      tmp = pow(tmp,2);
      gsl_vector_set (f, i, (tmp - y[i]));
    }
  return GSL_SUCCESS;
}

void
callback(const size_t iter, void *params,
         const gsl_multifit_nlinear_workspace *w)
{
    size_t n = ((struct data *)params)->n;
    gsl_vector *f = gsl_multifit_nlinear_residual(w);
    double fvalue = gsl_blas_dnrm2(f);
    fvalue *=fvalue;

    if (iter == 0)
          printf("Iteration   Function Value          Current  Solution\n");
    printf ("%5ld       %10.5f", iter,
            fvalue);

    size_t max_out = n;
    if (max_out>5) max_out=5;

    gsl_vector *x = gsl_multifit_nlinear_position(w);
    for (size_t i=0; i<max_out; i++){
        printf ("          %.5f",
                gsl_vector_get (x, i));
    }
    printf("\n");
}

int
main (int argc, char* argv[])
{
  assert(argc==3); size_t m=std::stoi(argv[1]), n=std::stoi(argv[2]);

  const gsl_multifit_nlinear_type *T = gsl_multifit_nlinear_trust;
  gsl_multifit_nlinear_workspace *w;
  gsl_multifit_nlinear_fdf fdf;
  gsl_multifit_nlinear_parameters fdf_params =
    gsl_multifit_nlinear_default_parameters();

  gsl_vector *f;

  Eigen::Matrix<double, Eigen::Dynamic, 1> y;
  Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> t;
  /* this is the data to be fitted */
  t = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(m,n);
  y = Eigen::Matrix<double, Eigen::Dynamic, 1>::Random(m);
  t=t.cwiseProduct(t);
  double *Y = new double[m];
  double *tt = new double[m*n];
  for (size_t i=0; i<m; i++) Y[i] = y(i);
  for (size_t i=0; i<m; i++){
      for (size_t j=0; j<n; j++)
          tt[i*n+j] = t(i,j);
  }

  struct data params = { m, n, tt, Y};

  double *x_init = new double[n];
  for (size_t i=0; i<n; i++) x_init[i] = -1.0; /* starting values */

  gsl_vector_view x = gsl_vector_view_array (x_init, n);

  double chisq, chisq0;
  int status, info;

  const double xtol = 1e-8;
  const double gtol = 1e-8;
  const double ftol = 0.0;

  /* define the function to be minimized */
  fdf.f = expb_f;
  fdf.df = NULL; //expb_df;   /* set to NULL for finite-difference Jacobian */
  fdf.fvv = NULL;     /* not using geodesic acceleration */
  fdf.n = m;
  fdf.p = n;
  fdf.params = &params;

  /* allocate workspace with default parameters */
  w = gsl_multifit_nlinear_alloc (T, &fdf_params, m, n);

  /* initialize solver with starting point */
  gsl_multifit_nlinear_init (&x.vector, &fdf, w);
  /* compute initial cost function */
  f = gsl_multifit_nlinear_residual(w);
  gsl_blas_ddot(f, f, &chisq0);

  /* solve the system with a maximum of 1000 iterations */
  status = gsl_multifit_nlinear_driver(1000, xtol, gtol, ftol,
                                       callback, &params, &info, w);

  /* compute final cost */
  gsl_blas_ddot(f, f, &chisq);

  fprintf(stderr, "summary from method '%s/%s'\n",
          gsl_multifit_nlinear_name(w),
          gsl_multifit_nlinear_trs_name(w));
  fprintf(stderr, "number of iterations: %zu\n",
          gsl_multifit_nlinear_niter(w));
  fprintf(stderr, "function evaluations: %zu\n", fdf.nevalf);
  fprintf(stderr, "Jacobian evaluations: %zu\n", fdf.nevaldf);
  if (info !=0)
      fprintf(stderr, "reason for stopping: %s\n",
              (info == 1) ? "small step size" : "small gradient");
  else
      fprintf(stderr, "The algorithm has not converged. See status for the reason.\n");
  fprintf(stderr, "initial |f(x)| = %f\n", sqrt(chisq0));
  fprintf(stderr, "final   |f(x)| = %f\n", sqrt(chisq));

  {

      printf ("Minimum found at:\n");

      printf ("%5ld       %10.5f", gsl_multifit_nlinear_niter(w),
              chisq);
      size_t max_out = n;
      if (max_out>5) max_out=5;

      for (size_t i=0; i<max_out; i++){
          printf ("          %.5f",
                  gsl_vector_get (w->x, i));
      }
      printf("\n");


  }
  fprintf (stderr, "status = %s\n", gsl_strerror (status));

  gsl_multifit_nlinear_free (w);

  delete [] x_init;
  delete [] Y;
  delete [] tt;

  return 0;
}
